#
#  ApplicationController.rb
#  CocoaNav
#
#  Created by Patrick Geiller on 29/02/08.
#  Copyright (c) 2008 __MyCompanyName__. All rights reserved.
#

require 'osx/cocoa'
require_framework 'QuartzCore'
require_framework 'ImageKit'
require_framework 'CoreData'

#
# Application Controller is the application delegate,
#	loads frameworks, their objects and methods as CNItems
#
#	creates new documents
#	computes the class tree
#	walks the class tree with dom like methods
#
#	setups default colors
#	handles colors
#
class ApplicationController < NSObject

	kvc_accessor :items
	kvc_accessor :numFrameworks
	
	# Shared used defaults
	ib_outlet :userDefaults
	
	# Preferences window
	kvc_accessor :preferencesWindow
	ib_outlet :preferencesWindow
	
	ib_outlet :frameworkColors

	#
	# Awake
	#
	def awakeFromNib
		# We copy the method name in table view
		NSPasteboard.generalPasteboard.declareTypes_owner(NSArray.arrayWithObjects(NSStringPboardType, nil), self)
	
		@numFrameworks = 0
		@items				= NSMutableArray.alloc.init
		@frameworkPaths		= []
		@frameworkOrigins	= {}

		NSColorPanel.sharedColorPanel.setShowsAlpha(true)


		setupDefaults
	end

	#
	#	Application just launched
	#
	def	applicationDidFinishLaunching(notif)
		path = '/Developer/SDKs/MacOSX10.5.sdk/System/Library/Frameworks'
		path = '/System/Library/Frameworks/'
		self.performSelector_withObject_afterDelay(:findAndLoadFrameworksFromPath, path, 0)
	end
	
	
	#
	#	findAndLoadFrameworks
	#		grab all frameworks from a path and parse them
	#
	def	findAndLoadFrameworksFromPath(path)
		# Find all entries containing 'framework'
		paths = Dir.entries(path).find_all { |d| d =~ /framework/ }.map { |d| path + '/' + d }
		
#		paths = [path + '/Foundation.framework']
#		paths = [path + '/Foundation1.framework', path + '/Foundation.framework' , path + '/Foundation.framework', path + '/Foundation.framework', path + '/Foundation.framework' , path + '/Foundation.framework', path + '/Foundation.framework', path + '/Foundation.framework', path + '/Foundation.framework' , path + '/Foundation.framework', path + '/Foundation.framework', path + '/Foundation.framework' , path + '/Foundation.framework', path + '/Foundation.framework', path + '/Foundation.framework', path + '/Foundation.framework', path + '/Foundation.framework', path + '/Foundation.framework', path + '/Foundation.framework', path + '/Foundation.framework']

#		paths = [path + 'XGridFoundation.framework', path + 'RubyCocoa.framework', path + 'Accelerate.framework', path + 'Automator.framework']

#		paths = [path + 'XGridFoundation.framework', path + 'RubyCocoa.framework', path + 'Automator.framework']
#		paths = [path + 'AppKit.framework', path + 'Foundation.framework']

#		paths = [path + 'XGridFoundation.framework', path + 'CoreData.framework']


=begin
		paths =	[	path + 'XGridFoundation.framework', 
					path + 'RubyCocoa.framework', 
					path + 'Foundation1.framework', 
					path + 'Foundation1.framework', 
					path + 'Foundation1.framework', 
					path + 'Foundation1.framework', 
					path + 'Automator.framework', 
					path + 'Foundation1.framework', 
					path + 'Foundation1.framework', 
					path + 'Foundation1.framework', 
		
				path + 'InputMethodKit.framework']
=end

#		paths = [path + 'XGridFoundation.framework', path + 'RubyCocoa.framework']
#		paths = [path + 'CoreData.framework', path + 'XGridFoundation.framework', path + 'RubyCocoa.framework']
#		paths = [path + 'AppKit.framework', path + 'XGridFoundation.framework', path + 'RubyCocoa.framework']
		startLoadingFrameworksDeferred(paths)
	end
	
	#
	#	startLoadFrameworksDeferred
	#		load each framework with a performSelector afterDelay
	#
	def	startLoadingFrameworksDeferred(paths)
		@currentFrameworkPaths = paths
		@currentFrameworkLoadIdx = 0
		
		setValue_forKey(paths.length, 'numFrameworks')
		@numFrameworks = paths.length
		loadFrameworkDeferred
	end

	#
	#	loadFrameworkDeferred
	#		load a framework from currentFrameworkPaths 
	#		and call itself again with performSelector to load remaining frameworks
	#
	def	loadFrameworkDeferred
		frameworkPath = @currentFrameworkPaths[@currentFrameworkLoadIdx]

		loadFramework(frameworkPath, 'sys')
		
		if @currentFrameworkLoadIdx < @currentFrameworkPaths.length-1
			@currentFrameworkLoadIdx += 1
			self.performSelector_withObject_afterDelay(:loadFrameworkDeferred, nil, 0)
		else
			loadedFrameworks
		end
	end

	#
	# Loaded frameworks
	#
	def loadedFrameworks
		# Add dummy item to signal we're done loading
		obj		= CNItem.alloc.init
		obj.name = ''
		obj.file = ''
		obj.type = 'end'
		obj.origin		= ''
		obj.searchName	= ''
		obj.framework	= ''
		addItem(obj)

		computeHierarchy
	end
	
	#
	#	loadFramework
	#		grab all framework files and parse classes, methods
	#
	def	loadFramework(frameworkPath, origin)
		name	= frameworkPath.to_s.split('/')[-1]
		name	= name.split('.framework')[0] if name.match(/\.framework/)

		@frameworkPaths << frameworkPath
		@frameworkOrigins[frameworkPath.to_s] = origin
		
		# Init item as framework
		obj		= CNItem.alloc.init
		obj.name = name
		obj.file = frameworkPath
		obj.type = 'framework'
		obj.origin		= origin
		obj.searchName	= obj.name
		obj.framework	= frameworkPath
		addItem(obj)

		files = `find '#{frameworkPath}' | grep -i \\\\.h$`.split("\n")
		files.each { |file| parseFile(file, frameworkPath, origin) }	
	end
	
	#
	# isFrameworkLoaded
	#
	def isFrameworkLoaded(frameworkPath)
		@frameworkPaths.index(frameworkPath) != nil
	end

	#
	#	parseFile
	#
	def	parseFile(file, frameworkPath, origin)
		# Bail when encountering non existing files (some aliases are pointing to nothing in Ruby.framework)
		if !File.exists?(file)
#			puts 'parseFile : #{file} not found'
			return
		end

		# Bail if file doesn't contain any class definition
		fileContent =  IO.read(file)
		return	if (fileContent =~ /@interface/).nil?
		
		# Parse classes 
		lines = fileContent.split("\n")
		i = 0
		while i < lines.length
			line = lines[i]
			if line =~ /^@interface/
				numLinesParsed = parseClass(lines, file, i, frameworkPath, origin)
				# Skip that line if we can't parse it
				numLinesParsed = 1 if numLinesParsed == 0
				i += numLinesParsed
			else
				i+= 1
			end
		end
	end
	
	#
	#	parseClass
	#		called when encountering lines containing @interface
	#		returns number of lines parse
	#
	def	parseClass(lines, file, i, frameworkPath, origin)
		# Parse class definition
		#
		# match @interface NSURLProtectionSpace : NSObject <NSCopying, NSTrucmuche>
		# to
		# 'NSURLProtectionSpace', ': NSObject', '<NSCopying, NSTrucmuche>'
		line = lines[i]
		scan = line.scan(/@interface\s*(\w+)\s*(:\s*\w+)?\s*(<[^>]+>)?(\(\s*\w+\s*\))?/)[0]
		className = scan[0].to_s
		parentClassName = scan[1].nil? ? nil : scan[1].match(/\w+/).to_s
		protocols = scan[2].nil? ? nil : scan[2].match(/[\w\s,]+/).to_s.split(',').collect { |x| x.strip }
		category = scan[3].nil? ? nil : scan[3].match(/\w+/).to_s
		
#		str = 'class=' + className
#		str += ' parent=' + (parentClassName||'')
#		str += ' category=' + category if !category.nil?
#		str += ' protocols=' + protocols.to_s if !protocols.nil?
#		puts lines[i]
#		puts '  ' + str
#		puts className + '.' + frameworkPath

		# Class
		if category.nil?
			# Init item as class
			obj = CNItem.alloc.init
			obj.name = className
			obj.type = 'class'
			obj.file = file
			obj.line = i
			obj.searchName	= obj.name
			obj.framework	= frameworkPath
			obj.parentClass = parentClassName
			obj.origin		= origin
			addItem(obj)
		else
		# Category
			obj = CNItem.alloc.init
			obj.name = category
			obj.type = 'category'
			obj.file = file
			obj.line = i
			obj.searchName	= 'category ' + obj.name + ' ' + className
			obj.framework	= frameworkPath
			obj.origin		= origin
			obj.associatedClass = className
			addItem(obj)
		end
		# Protocols
		if protocols != nil
			protocols.each { |protocol|
				obj = CNItem.alloc.init
				obj.name = protocol
				obj.type = 'protocol'
				obj.file = file
				obj.line = i
				obj.searchName	= 'protocol ' + obj.name + ' ' + className
				obj.framework	= frameworkPath
				obj.origin		= origin
				obj.associatedClass = className
				addItem(obj)
			}
		end

		j = i + 1
		# Parse class methods
		while j < lines.length
			line = lines[j]
			if line =~ /@end/
				return j-i
			end
			
			# If line starts with a + or -, we're parsing a class or instance method
			if line =~ /^(\+|\-)\s+/
				parseMethod(className, file, line, j, frameworkPath, origin)
			end
			
			j += 1
		end
		
		# We didn't parse anything
		0
	end
	
	
	#
	#	parseMethod
	#		called
	def	parseMethod (className, file, line, i, frameworkPath, origin)
		# If method has arguments, parse it as words and join fragments with ':'
		if line =~ /:/
			name = line.scan(/(\w+):/).join(':') + ':'
		else
			name = line.scan(/(\w+);/).join('')
		end

		obj = CNItem.alloc.init
		obj.name = name
		obj.type = 'method'
		obj.file = file
		obj.line = i
		obj.searchName		= obj.name
		obj.framework		= frameworkPath
		obj.associatedClass = className
		obj.isInstanceMethod= line.match(/-/)
		obj.origin			= origin
		addItem(obj)
	end


	#
	# Add item, KVO compliant 
	#
	def addItem(obj)
		insertedIndexes = NSIndexSet.indexSetWithIndex(@items.count)
		willChange_valuesAtIndexes_forKey(NSKeyValueChangeInsertion, insertedIndexes, 'items');
		@items.addObject obj
		didChange_valuesAtIndexes_forKey(NSKeyValueChangeInsertion, insertedIndexes, 'items');
	end



	#
	# New document
	#	set 'default' type - load all frameworks
	#
	def newDocument(sender)
		# custom forwarding
		# http://developer.apple.com/documentation/Cocoa/Conceptual/EventOverview/EventArchitecture/chapter_2_section_6.html
		NSDocumentController.sharedDocumentController.newDocument(sender)
		NSDocumentController.sharedDocumentController.currentDocument.setValue_forKey('default', 'type')
	end
	
	#
	# New blank document
	#	set 'blank' type - load nothing
	#
	def newBlankDocument(sender)
		NSDocumentController.sharedDocumentController.newDocument(sender)
		NSDocumentController.sharedDocumentController.currentDocument.setValue_forKey('blank', 'type')
	end
	
	
	
	#
	# Precompute some data 
	#
	#	hashParent		class name => parent class name
	#	hashChildren	class name => array of classname's children
	#	rootObjects		parentless classes
	#	hashFramework	class name => framework
	#
	def	computeHierarchy
		# Compute hashParent and hashChildren
		@hashParent		= {}
		@hashChildren	= {}
		@hashFramework	= {}
		@items.each { |item|
			if item.oc_type == 'class'
				@hashParent[item.name.to_s] = item.parentClass.to_s
				if @hashChildren[item.parentClass.to_s] == nil
					@hashChildren[item.parentClass.to_s] = [item.name.to_s]
				else
					@hashChildren[item.parentClass.to_s] << item.name.to_s
				end
				@hashFramework[item.name.to_s] = item.framework
			end
		}
		
		# Sort children
		@hashChildren.each_pair { |key, value|
			value = value.sort!
		}
		
		# Compute rootObjects
		@rootObjects = []
		@hashParent.each_pair { |key, value|
			# Add objects who have no parents
			@rootObjects << key if (value == nil || value == '') && !@rootObjects.include?(key)

			# AND add objects whose parents have no parents
			# This is useful to create top level entries when loading just one framework, 
			# as NSView and the like may not be loaded
			if !@hashParent.has_key?(value) && value.length > 0
				@rootObjects << value if !@rootObjects.include?(value)
			end
		}
	end
	# Precomputation accessors
	def hashChildren
		computeHierarchy if @hashChildren == nil
		return	@hashChildren
	end
	def parentClassOfClass(c)
		computeHierarchy if @hashParent == nil
		return	@hashParent[c.to_s]
	end
	def rootObjects
		computeHierarchy if @rootObjects == nil
		return	@rootObjects
	end





	#
	#
	# Helper methods to move around classes, DOM-like : parent, first child, last child, previous sibling, next sibling, next, previous
	#
	#

	#
	# parent
	#
	def	classParent(c)
		NSApp.delegate.parentClassOfClass(c.to_s)
	end
	
	#
	# first parent
	#
	def classRootOf(c)
		while c
			d = classParent(c)
			return c if d == nil
			c = d
		end
	end
	
	#
	# next sibling
	#
	def classNextSibling(c)
		v = NSApp.delegate.parentClassOfClass(c.to_s)
		return nil if v == nil
		children = NSApp.delegate.hashChildren[v.to_s]
		i = children.index(c.to_s)
		i += 1
		return if i >= children.length
		children[i]
	end
	
	#
	# previous sibling
	#
	def classPreviousSibling(c)
		v = NSApp.delegate.parentClassOfClass(c.to_s)
		return nil if v == nil
		children = NSApp.delegate.hashChildren[v.to_s]
		i = children.index(c.to_s)
		i -= 1
		return if i < 0
		children[i]
	end

	#
	# next sibling wrap : go back to first sibling
	#
	def classNextSiblingWrap(c)
		v = NSApp.delegate.parentClassOfClass(c.to_s)
		return nil if v == nil
		children = NSApp.delegate.hashChildren[v.to_s]
		i = children.index(c.to_s)
		i += 1
		i = 0 if i >= children.length
		children[i]
	end
	
	#
	# previous sibling wrap : go to last sibling
	#
	def classPreviousSiblingWrap(c)
		v = NSApp.delegate.parentClassOfClass(c.to_s)
		return nil if v == nil
		children = NSApp.delegate.hashChildren[v.to_s]
		i = children.index(c.to_s)
		i -= 1
		i = children.length-1 if i < 0
		children[i]
	end

	#
	# first child
	#
	def classFirstChild(c)
		children = NSApp.delegate.hashChildren[c.to_s]
		return nil if children == nil
		children[0]
	end
	
	#
	# last child
	#
	def classLastChild(c)
		children = NSApp.delegate.hashChildren[c.to_s]
		return nil if children == nil
		children[-1]
	end
	
	#
	# next
	#	gives the next node in the hierarchy. If the classes were displayed in an NSOutlineView, 
	#	this would be equivalent to getting the class of the line below
	#
	def	classNext(c)
		
		f = classFirstChild(c)
		return f if f
		
		n = c
		while n
			n2 = classNextSibling(n)
			return n2 if n2
			n = classParent(n)
		end
		
		# Nothing found, go back to root and take its next sibling in rootObjects
		root = classRootOf(c)
		rootObjects = NSApp.delegate.rootObjects
		idx = rootObjects.index(root)
		idx += 1
		idx = 0 if idx >= rootObjects.length
		return rootObjects[idx]
	end

	#
	# previous
	#	the 'line above'
	#
	def	classPrevious(c)
		if classPreviousSibling(c.to_s)
			n = classPreviousSibling(c.to_s)
			while n && classLastChild(n)
				n = classLastChild(n)
			end
			return n
		end
		p = classParent(c.to_s)
		return p if p
		
		# Nothing found, go back to root and take its previous sibling in rootObjects
		root = classRootOf(c)
		rootObjects = NSApp.delegate.rootObjects
		idx = rootObjects.index(root)
		idx -= 1
		idx = rootObjects.length-1 if idx < 0
		c = rootObjects[idx]

		# Take ultimate last child
		while c
			d = classLastChild(c)
			return c if d == nil
			c = d
		end		
	end
	
	
	#
	# Color for framework
	#
	def	colorForFramework(frameworkPath)
		return	CGColorCreateGenericRGB(1.0, 1.0, 1.0, 0.0) if frameworkPath == nil
		# Default color
		color = @userDefaults.values.valueForKey(@frameworkOrigins[frameworkPath.to_s] == 'sys' ? 'systemFrameworkColor' : 'userFrameworkColor')

		# Check custom colors
		frameworkName = frameworkPath.split('/')[-1].split('.')[0]
		@frameworkColors.arrangedObjects.each { |frameworkColor|
			next if	!frameworkName.to_s.match(Regexp.new(frameworkColor.name, Regexp::IGNORECASE)) 			
			next if frameworkColor.enabled == 0 
			next if frameworkColor.color == nil
			color = frameworkColor.color
		}
		return	archivedNSColorToCGColor(color)
	end


	#
	# Text color for framework
	#
	def	textColorForFramework(frameworkPath)
		return	CGColorCreateGenericRGB(1.0, 1.0, 1.0, 1.0) if frameworkPath == nil
		# Default color
		color = @userDefaults.values.valueForKey(@frameworkOrigins[frameworkPath.to_s] == 'sys' ? 'systemFrameworkTextColor' : 'userFrameworkTextColor')

		# Check custom colors
		frameworkName = frameworkPath.split('/')[-1].split('.')[0]
		@frameworkColors.arrangedObjects.each { |frameworkColor|
			next if	!frameworkName.to_s.match(Regexp.new(frameworkColor.name, Regexp::IGNORECASE)) 
			next if frameworkColor.enabled == 0 
			next if frameworkColor.textColor == nil
			color = frameworkColor.textColor
		}
		return	archivedNSColorToCGColor(color)
	end
	
	#
	# archived NSColor (defaults, Core Data)-> CGColor (Core Animation)
	#
	def	archivedNSColorToCGColor(color)
		color = NSValueTransformer.valueTransformerForName(NSUnarchiveFromDataTransformerName).transformedValue(color)
		color = color.colorUsingColorSpaceName(NSCalibratedRGBColorSpace)
		return	CGColorCreateGenericRGB(color.redComponent, color.greenComponent, color.blueComponent, color.alphaComponent)
	end

	#
	# NSColor to archived NSColor
	#
	def	NSColorToArchivedColor(color)
		NSValueTransformer.valueTransformerForName(NSUnarchiveFromDataTransformerName).reverseTransformedValue(color)
	end
	
	
	#
	# Called from preferences window when colors changed
	#	We set a dummy value observed by document windows, they'll update their views accordingly
	#
	kvc_accessor :colorChanged
	def colorChanged
		setValue_forKey(CFAbsoluteTimeGetCurrent(), 'colorChanged')
	end
	
	#
	# Given a class, return its framework
	#
	def	frameworkOfClass(className)
		@hashFramework[className.to_s]	
	end

	
	#
	# Setup defaults in NSUserDefaultsController
	#
	def setupDefaults
		defaults = NSUserDefaults.standardUserDefaults
		
#		backgroundColor1 = NSColorToArchivedColor(NSColor.colorWithDeviceRed_green_blue_alpha(0.07, 0.19, 0.37, 1))
#		backgroundColor2 = NSColorToArchivedColor(NSColor.colorWithDeviceRed_green_blue_alpha(0.57, 0.76, 1.00, 1))
		backgroundColor1 = NSColorToArchivedColor(NSColor.colorWithDeviceRed_green_blue_alpha(0.09, 0.11, 0.30, 1))
		backgroundColor2 = NSColorToArchivedColor(NSColor.colorWithDeviceRed_green_blue_alpha(0.62, 0.79, 1.00, 1))

		systemFrameworkColor = NSColorToArchivedColor(NSColor.colorWithDeviceRed_green_blue_alpha(0, 0, 0, 0))
		systemFrameworkTextColor = NSColorToArchivedColor(NSColor.colorWithDeviceRed_green_blue_alpha(1, 1, 1, 1))

		userFrameworkColor = NSColorToArchivedColor(NSColor.colorWithDeviceRed_green_blue_alpha(0.21, 0.57, 0.91, 0.8))
		userFrameworkTextColor = NSColorToArchivedColor(NSColor.colorWithDeviceRed_green_blue_alpha(0, 0, 0, 0.8))

		values = NSDictionary.dictionaryWithObjectsAndKeys(	backgroundColor1, 'backgroundColor1', 
															backgroundColor2, 'backgroundColor2', 
															systemFrameworkColor, 'systemFrameworkColor',
															systemFrameworkTextColor, 'systemFrameworkTextColor',
															userFrameworkColor, 'userFrameworkColor',
															userFrameworkTextColor, 'userFrameworkTextColor',
															true, 'firstRun',
															nil)
		defaults.registerDefaults(values)
		isFirstRun = defaults.boolForKey('firstRun')
		defaults.setBool_forKey(false, 'firstRun')

		return	if !isFirstRun
		# Register framework colors on first run
		addFrameworkColor('(AppKit$|^Cocoa$|^Foundation$)', NSColor.colorWithDeviceRed_green_blue_alpha(0, 0, 0, 0.85), NSColor.whiteColor)
		addFrameworkColor('WebKit', NSColor.colorWithDeviceRed_green_blue_alpha(0.82, 1, 0.05, 0.85), NSColor.blackColor)
		addFrameworkColor('Sync', NSColor.colorWithDeviceRed_green_blue_alpha(0.31, 0.76, 1, 0.85), NSColor.whiteColor)
		addFrameworkColor('Quartz', NSColor.colorWithDeviceRed_green_blue_alpha(0.14, 0.24, 1, 0.85), NSColor.whiteColor)
		addFrameworkColor('CoreData', NSColor.colorWithDeviceRed_green_blue_alpha(1, 0.56, 0, 0.85), NSColor.whiteColor)
	end
	
	#
	# Add a default FrameworkColor (Core Data managed objects)
	#
	def	addFrameworkColor(frameworkName, color, textColor)
		frameworkColorClass = managedObjectModel.entitiesByName['FrameworkColor']
		o = NSManagedObject.alloc.initWithEntity_insertIntoManagedObjectContext(frameworkColorClass, managedObjectContext)
		o.setValue_forKey(frameworkName, "name")
		o.setValue_forKey(NSColorToArchivedColor(color), "color")
		o.setValue_forKey(NSColorToArchivedColor(textColor), "textColor")
	end
	


	#
	#
	#	Core Data
	#
	#


  def managedObjectModel
    return @managedObjectModel if @managedObjectModel
    
    allBundles = OSX::NSMutableSet.alloc.init
    allBundles.addObject(OSX::NSBundle.mainBundle)
    allBundles.addObjectsFromArray(OSX::NSBundle.allFrameworks)
    
    @managedObjectModel = OSX::NSManagedObjectModel.mergedModelFromBundles(allBundles.allObjects)
    OSX::CoreData.define_wrapper(@managedObjectModel)
    return @managedObjectModel
  end
  
  # Change this path/code to point to your App's data store.
  def applicationSupportFolder
    OSX.NSHomeDirectory.stringByAppendingPathComponent(File.join('Library', 'Application Support', 'CocoaNav'))
  end
  
  def managedObjectContext
    return @managedObjectContext if @managedObjectContext
    
    fileManager = OSX::NSFileManager.defaultManager
    storeFolder = applicationSupportFolder
    unless fileManager.fileExistsAtPath_isDirectory?(storeFolder, nil)
      fileManager.createDirectoryAtPath_attributes(storeFolder, nil)
    end
    
    url = OSX::NSURL.fileURLWithPath(storeFolder.stringByAppendingPathComponent('CocoaNav.xml'))
    coordinator = OSX::NSPersistentStoreCoordinator.alloc.initWithManagedObjectModel(managedObjectModel)
    # FIXME: cannot get errors
    error = OSX::OCObject.new
    if coordinator.addPersistentStoreWithType_configuration_URL_options_error(OSX::NSXMLStoreType, nil, url, nil, error) then
      @managedObjectContext = OSX::NSManagedObjectContext.alloc.init
      @managedObjectContext.setPersistentStoreCoordinator(coordinator)
    else
      OSX::NSApplication.sharedApplication.presentError(error)
    end
    
    return @managedObjectContext
  end

  def windowWillReturnUndoManager(window)
    return managedObjectContext.undoManager
  end
  
  def saveAction(sender)
    error = nil
    unless managedObjectContext.save?(error)
      OSX::NSApplication.sharedApplication.presentError(error)
    end
  end
  ib_action :saveAction
  
  def applicationShouldTerminate(sender)
    reply = OSX::NSTerminateNow
#    return true
    context = managedObjectContext
    if context
      if context.commitEditing?
        # FIXME: cannot get errors
        error = OSX::OCObject.new
        unless context.save?(error)
          errorResult = OSX::NSApplication.sharedApplication.presentError?(error)
        
          if errorResult
            reply = OSX::NSTerminateCancel
          else
            alertReturn = OSX.NSRunAlertPanel(nil, 'Could not save changes while quitting. Quit anyway?', 'Quit anyway', 'Cancel', nil)
            if alertReturn == OSX::NSAlertAlternateReturn
              reply = OSX::NSTerminateCancel
            end
          end
        end
      else
        reply = OSX::NSTerminateCancel
      end
    end
    return reply
  end
end




