#
#  FrameworkView.rb
#  CocoaNav
#
#  Created by Patrick Geiller on 02/04/08.
#  Copyright (c) 2008 __MyCompanyName__. All rights reserved.
#

#
# FrameworkView draws Frameworks and Classes in two possible views : by Framework, or in a Class Tree
#
class FrameworkView < FrameworkViewBackground


	#
	# Class Tree
	#
	def	awakeFromNib

		# Setup Core Animation
		setWantsLayer true
		@rootLayer = self.layer
		
		# frameworksLayer contains all our stuff
		@frameworksLayer = CALayer.layer
		@frameworksLayer.position = CGPointMake(frame.width/2, frame.height/2)
		@rootLayer.addSublayer @frameworksLayer

		# Used to get framework layer coordinates
		@helperLayer = CALayer.layer
		@frameworksLayer.addSublayer @helperLayer

		# Sun layer
		qtzPath = NSBundle.mainBundle.pathForResource_ofType('Sun', 'qtz')
		@sunLayer = QCCompositionLayer.compositionLayerWithFile(qtzPath)
		layer.addSublayer @sunLayer
		@sunLayer.bounds = CGRectMake(0, 0, frame.width, frame.height)
		@sunLayer.anchorPoint = CGPointMake(0, 0)

		# gradient image backing framework layers 
		path = NSBundle.mainBundle.pathForImageResource('gradient2.png')
		provider = CGDataProviderCreateWithURL(NSURL.alloc.initFileURLWithPath(path))
		@image = CGImageCreateWithPNGDataProvider(provider, nil, true, KCGRenderingIntentDefault)
		CGDataProviderRelease(provider)



		@frameworks = []
		
		@startZPosition = -200
		@launchZPosition = 150
		@spaceBetweenLayers = 100

		@cameraMaximumZPosition = 50
		@zDistance = 450

		@numColumns = 6
		@frameworkLayerWidth = 800
		@frameworkLayerStartHeight = 140
		@classLayerWidth = @frameworkLayerWidth / (@numColumns)



		@classLayerFontSize = 10
		
		
		@frameworksLayerX = @frameworksLayerY = @frameworksLayerZ = 0
		updateFrameworksLayerTransform(@frameworksLayerX, @frameworksLayerY, @frameworksLayerZ)
		@frameworksLayer.sublayerTransform = CATransform3DRotate(@frameworksLayer.sublayerTransform, 0.3, 1, 0, 0)

		
		@dragging = false
		
		@zoomAmount = 600

		# Default colors
		@defaultFrameworkTextColor = CGColorCreateGenericRGB(0, 0, 0, 1)
		@defaultFrameworkColor = CGColorCreateGenericRGB(0.9, 0.9, 0.9, 0.8)
		@defaultFrameworkColor = CGColorCreateGenericRGB(0.9, 0.9, 0.9, 0.9)
		@defaultFrameworkColor = CGColorCreateGenericRGB(1, 1, 1, 0.8)
		@frameworkTitleBackgroundColor = CGColorCreateGenericRGB(0, 0, 0, 0.08)
		@frameworkTitleBackgroundColor = CGColorCreateGenericRGB(0, 0, 0, 0.48)
		@defaultFrameworkWireColor = CGColorCreateGenericRGB(1, 1, 1, 0.2)
		
		
		# Accept files
		self.registerForDraggedTypes(NSArray.arrayWithObjects(NSFilenamesPboardType, nil))
		
		
		@viewMode = 'framework'
	
		@loadedFrameworks = false

		# the yellow selection layer, overlaid the selected class layer
		createSelectionLayer

		# Used to refresh colors in view switch
		@colorsChangedSinceLastViewSwitch = false

		# Selected class
		@value = ''

		@isFirstResponder = false
		
		@zPositionWhenSelecting = -200
	end
	
	
	
	#
	# New framework layer
	#
	def	addFramework(frameworkPath, origin)
		animateFrameworks if @loadedFrameworks == false

		# Get colors
		@frameworkColor		= colorForFramework(frameworkPath)
		@frameworkTextColor	= textColorForFramework(frameworkPath)

		frameworkName = frameworkPath.split('/')[-1].split('.')[0]
		# framework background layer
		frameworkLayer = CALayer.layer
		width = @frameworkLayerWidth
		height = @frameworkLayerStartHeight
		frameworkLayer.bounds = CGRectMake(0, 0, width, height)
		frameworkLayer.backgroundColor = @frameworkColor
		frameworkLayer.contents = @image
		frameworkLayer.anchorPoint = CGPointMake(0.5, 1)

		# framework title layer
		titleLayer = CATextLayer.layer
		titleLayer.fontSize = 60
		titleLayer.foregroundColor = @frameworkTextColor
		titleLayer.backgroundColor = @frameworkTitleBackgroundColor
		titleLayer.cornerRadius = titleLayer.fontSize/2
		titleLayer.position = CGPointMake(frameworkLayer.bounds.size.width/2, height-titleLayer.fontSize)
		titleLayer.string = frameworkName
		titleLayer.style = NSDictionary.dictionaryWithObjectsAndKeys('Futura-MediumItalic', 'font', KCAAlignmentCenter, 'alignmentMode', nil)
		s = titleLayer.preferredFrameSize
		titleLayer.bounds = CGRectMake(0, 0, s.width+titleLayer.fontSize, s.height)

		titleLayer2 = CATextLayer.layer
		titleLayer2.string = frameworkPath

		frameworkLayer.addSublayer titleLayer
		titleLayer.addSublayer titleLayer2
		@frameworksLayer.addSublayer frameworkLayer

		@lastFrameworkLayer = frameworkLayer
		if @loadedFrameworks == false
			CATransaction.setValue_forKey(0, KCATransactionAnimationDuration)
			frameworkLayer.zPosition = @launchZPosition
			CATransaction.commit
			frameworkLayer.zPosition = @startZPosition
			CATransaction.commit
		else
		
		end
		
		@frameworks << @lastFrameworkLayer
		@textLayerX = 0
		@textLayerY = 12
	end
	




	#
	# New class layer
	#
	def	addClass(className)
		layer = CATextLayer.layer
		layer.fontSize = @classLayerFontSize
		layer.foregroundColor = @frameworkTextColor
		layer.position = CGPointMake(@lastFrameworkLayer.bounds.size.width/2, -@lastFrameworkLayer.bounds.size.height*2.7)
		layer.string = className
		layer.style = NSDictionary.dictionaryWithObjectsAndKeys('Futura-MediumItalic', 'font', KCAAlignmentLeft, 'alignmentMode', nil)

		s = layer.preferredFrameSize
		layer.bounds = CGRectMake(0, 0, s.width, s.height)
		scale = 0.001
		scale = 2
		layer.transform = CATransform3DMakeScale(scale, scale, scale)
		layer.opacity = 0
		
		@lastFrameworkLayer.addSublayer layer

		CATransaction.commit

		layer.transform = CATransform3DMakeRotation(0, 1, 0, 0)
		layer.opacity = 1
		layer.anchorPoint = CGPointMake(0, 0)
		layer.position = CGPointMake(@textLayerX, @textLayerY)
		CATransaction.commit
	
		@textLayerX += @classLayerWidth
		if (@textLayerX+@classLayerWidth) > @frameworkLayerWidth
			@textLayerX = 0
			@textLayerY -= @classLayerFontSize
			
			@lastFrameworkLayer.bounds = CGRectMake(0, @frameworkLayerStartHeight, @frameworkLayerWidth, @textLayerY-@frameworkLayerStartHeight)
		end
	end
	

	#
	# Loaded frameworks
	#	zoom extent
	#
	def	loadedFrameworks
		CATransaction.begin
		CATransaction.setValue_forKey(2.8, KCATransactionAnimationDuration)
#		CATransaction.setValue_forKey(0, KCATransactionAnimationDuration)
		@sunLayer.opacity = 0
		
		layoutFrameworks
		zoomExtentLayers
		CATransaction.commit

		@loadedFrameworks = true
	end

	#
	# When loading, animate frameworks from front to rear
	#
	def	animateFrameworks
		if (@lastFrameworkLayer)
			@frameworks.each_with_index { |fw, i|
				z = -(@frameworks.length-i)*@spaceBetweenLayers+@startZPosition
				fw.zPosition = z
			}
		end
	
	end


	#
	# Layout frameworks in a grid respecting the view's aspect ratio
	#
	def	layoutFrameworks
		return	if @frameworks == nil || @frameworks.length == 0

		totalHeight = 0; @frameworks.each { |fw| totalHeight += fw.bounds.size.height }
		averageFrameworkLayerHeight = totalHeight / @frameworks.length
		
		# View's aspect ratio we have to match
		aspectRatio = frame.width / frame.height

		# Iteratively test possible number of columns
		closestDistance = 1e99
		bestMatchingNumberOfColumns = -1
		numberOfColumns = 1
		while numberOfColumns < 20
			w = numberOfColumns*@frameworkLayerWidth
			h = (@frameworks.length / numberOfColumns)*averageFrameworkLayerHeight
			
			r = w / h

			distance = (aspectRatio-r).abs
			if (distance < closestDistance)
				closestDistance = distance
				bestMatchingNumberOfColumns = numberOfColumns
			end
			numberOfColumns += 1
		end
		
		averageColumnHeight = (@frameworks.length/bestMatchingNumberOfColumns).ceil*averageFrameworkLayerHeight

		# Layout layers
		hpadding = 40
		vpadding = hpadding
		x = @frameworkLayerWidth/2
		y = averageColumnHeight
		paddedY = y
		previousHeight = 0
		@frameworks.each_with_index { |fw, i|
			y -= previousHeight
			paddedY -= previousHeight + vpadding
			previousHeight = fw.bounds.size.height
			
			if (y-previousHeight < 0)
				y = averageColumnHeight
				paddedY = y - vpadding
				x += @frameworkLayerWidth + hpadding
			end
			
			fw.position = CGPointMake(x, paddedY)
			fw.zPosition = 0
		}

		# Update selection position
		reflectSelectionInSelectionLayer
	end
	
	#
	# Compute extent of a layer's sublayers 
	#
	def	extentOfLayerInParentLayer(parentLayer)
		x1 = y1 = 1e99
		x2 = y2 = 0
		parentLayer.sublayers.each { |fw|
			# Using match instead of layer.class == 'CATextLayer' because layer.class is in fact 'OSX::CATextLayer'
			next if @viewMode == 'tree' and !fw.class.to_s.match(/Text/)
			next if fw == @selectionLayer
			x1 = [x1, fw.frame.origin.x].min
			y1 = [y1, fw.frame.origin.y].min
			x2 = [x2, fw.frame.origin.x+fw.frame.size.width].max
			y2 = [y2, fw.frame.origin.y+fw.frame.size.height].max
		}
#		puts 'abs=' + x1.to_s + ' *** ' + y1.to_s + ' *** ' + x2.to_s + ' *** ' + y2.to_s
		NSMakeRect(x1, y1, x2-x1, y2-y1)
	end
	
	#
	# Zoom extent
	#
	def zoomExtentLayers
		rect = extentOfLayerInParentLayer(@frameworksLayer)
#		puts 'rect=' + rect.origin.x.to_s + ' *** ' + rect.origin.y.to_s + ' *** ' + rect.size.width.to_s + ' ***' + rect.size.height.to_s
#		puts 'x=' + @frameworksLayerX.to_s + ' y=' + @frameworksLayerY.to_s

		# Recenter our container layer
		@frameworksLayer.anchorPoint = CGPointMake(0.5, 0.5)
		@frameworksLayer.position = CGPointMake(self.frame.width/2, self.frame.height/2)

		@frameworksLayerX = -(rect.origin.x*2+rect.size.width)/2
		@frameworksLayerY = -(rect.origin.y*2+rect.size.height)/2

		# Iteratively find z for extent
		i = 0
		left = -100000
		right = 0 
		# Don't go for more than 20 iterations
		while i < 20
			mid = (left+right)/2
			updateFrameworksLayerTransform(@frameworksLayerX, @frameworksLayerY, mid)
			screenRect = @helperLayer.convertRect_toLayer(CGRectMake(-rect.size.width/2, -rect.size.height/2, rect.size.width, rect.size.height), self.layer)

			v = screenRect.size.height / frame.height > screenRect.size.width / frame.width ? screenRect.size.height > frame.height : screenRect.size.width > frame.width
			if v
				right = mid
			else
				left  = mid
			end
			i += 1
		end
		@frameworksLayerZ = mid - 50 # some padding
		updateFrameworksLayerTransform(@frameworksLayerX, @frameworksLayerY, @frameworksLayerZ)
	end
	
	
	
	

	
	#
	# Resize only when loading to recenter layers
	#
	def viewDidEndLiveResize
		return	if @loadedFrameworks
		@frameworksLayer.position = CGPointMake(frame.width/2, frame.height/2)
		@sunLayer.bounds = CGRectMake(0, 0, frame.width, frame.height)
	end

	#
	# Context menu (go to parent, next sibling …) populate children list
	#
	def	rightMouseDown(event)
		return	if !@loadedFrameworks
	
		clickedLayer = self.layer.hitTest(NSPointToCGPoint(convertPoint_fromView(event.locationInWindow, nil)))
		@rightClickedLayer = isLayerAClassLayer(clickedLayer)

		# Update context menu
		# Get object list to display
		items = @source.hashChildren[@rightClickedLayer.to_s]

		# Reset menu by removing any previously inserted class
		m = menu.itemAtIndex(3).submenu
		while m.numberOfItems > 1 && m.itemAtIndex(1)
			m.removeItemAtIndex(1)
		end
		
		# Populate menu
		items = [] if items == nil
		items.each { |item|
			item = NSMenuItem.alloc.initWithTitle_action_keyEquivalent(item, :viewGoToChild, '')
			m.addItem item		
		}
		
		# Show or hide 'No children' menu item
		m.itemAtIndex(0).setHidden(items.length > 0)
		
		super_rightMouseDown(event)
	end
	
	#
	# Validate menu
	#	validate items if we clicked a layer on right mouse down
	#
	def validateMenuItem(item)
		return @rightClickedLayer if item.menu == menu
		return true
	end
	
	
	
	#
	#
	# Handling move and zoom
	#
	#

	def	frameworksLayerCoordinatesFromViewCoordinates(point)
		point2 = self.layer.convertPoint_toLayer(point, @helperLayer)
		return point2
	end
	
	
	def deltaDrag(position)
		positionInFrameworksLayerCoordinates1 = frameworksLayerCoordinatesFromViewCoordinates(position)
		positionInFrameworksLayerCoordinates2 = frameworksLayerCoordinatesFromViewCoordinates(CGPointMake(position.x+1, position.y))
		delta = CGPointMake((position.x - @dragStartPosition.x)*(positionInFrameworksLayerCoordinates2.x-positionInFrameworksLayerCoordinates1.x), (position.y - @dragStartPosition.y)*(positionInFrameworksLayerCoordinates2.x-positionInFrameworksLayerCoordinates1.x))
	end
	
	def	mouseDown(event)
		return	if !@loadedFrameworks

		@dragging = true
		@dragStartPosition = NSPointToCGPoint(convertPoint_fromView(event.locationInWindow, nil))
		@dragged = false
	end

	def	mouseDragged(event)
		# The view receives fantom mouseDragged when dragging the title bar and jerking the window around ...
		# Don't go in here unless we've been through mouseDown.
		return	if @dragging == false
		@dragged = true
		
		delta = deltaDrag(NSPointToCGPoint(convertPoint_fromView(event.locationInWindow, nil)))
		
		CATransaction.begin
		CATransaction.setValue_forKey(0.1, KCATransactionAnimationDuration)
		updateFrameworksLayerTransform(@frameworksLayerX + delta.x, @frameworksLayerY + delta.y, @frameworksLayerZ)
		CATransaction.commit
		
	end
	
	def	mouseUp(event)
		# If we didn't drag, select clicker layer
		if !@dragged
			clickedLayer = self.layer.hitTest(NSPointToCGPoint(convertPoint_fromView(event.locationInWindow, nil)))
			maybeSelectLayer(clickedLayer)
			return
		end

		return	if @dragging == false
		@dragging = false

		delta = deltaDrag(NSPointToCGPoint(convertPoint_fromView(event.locationInWindow, nil)))
		
		@frameworksLayerX += delta.x
		@frameworksLayerY += delta.y
	end
	
	def scrollWheel(event)
		return	if !@loadedFrameworks
	
		positionInView = NSPointToCGPoint(convertPoint_fromView(event.locationInWindow, nil))
		positionInLayer = frameworksLayerCoordinatesFromViewCoordinates(positionInView)
		
		if @lastScrollPosition == nil
			@lastScrollPosition = CGPointMake(-1, -1)
		end
		
		#
		# Recenter layer transform on (0, 0) for easy zooming. This means changing both transform AND position, to offset the difference.
		#
		# Recenter only when mouse position changed.
		samePositionAsLastScroll = @lastScrollPosition.x == positionInView.x && @lastScrollPosition.y == positionInView.y
		if samePositionAsLastScroll == false
			CATransaction.begin
			CATransaction.setValue_forKey(0, KCATransactionAnimationDuration)


			layerPosition = @frameworksLayer.position
			x = layerPosition.x + (positionInView.x-layerPosition.x)
			y = layerPosition.y + (positionInView.y-layerPosition.y)
			@frameworksLayer.position = CGPointMake(x, y)
			
			@frameworksLayerX = -positionInLayer.x
			@frameworksLayerY = -positionInLayer.y
			updateFrameworksLayerTransform(@frameworksLayerX, @frameworksLayerY, @frameworksLayerZ)
			CATransaction.commit

			
			@lastScrollPosition = positionInView
			return
		end
	
		s = @frameworksLayerZ
		s += event.deltaY*100
		s = [s, @cameraMaximumZPosition].min
		@frameworksLayerZ = s
	
		CATransaction.begin
		CATransaction.setValue_forKey(0.5, KCATransactionAnimationDuration)
		updateFrameworksLayerTransform(@frameworksLayerX, @frameworksLayerY, @frameworksLayerZ)
		CATransaction.commit
	end
	
	
	#
	# Update global transform with new position
	#
	def	updateFrameworksLayerTransform(x, y, z)
		transform = CATransform3DIdentity
		transform.m34 = 1.0 / -@zDistance
		transform = CATransform3DTranslate(transform, x, y, z)

		@frameworksLayer.sublayerTransform = transform
	end
	
	
	#
	#
	# Drag : allow drag of frameworks
	#
	#
	def draggingEntered(sender)
		return	NSDragOperationNone	if !@loadedFrameworks
		return	NSDragOperationCopy
	end
	
	# Drop
	def performDragOperation(sender)
		sender.draggingPasteboard.propertyListForType(NSFilenamesPboardType).each { |file|
			window.document.maybeLoadUserFramework(file)
			zoomExtent
		}
		return	true
	end
	
	
	def setSource(source)
		@source = source
	end
	

	#
	#
	# IBActions
	#
	#
	
	# Context menu
	ib_action :viewGoToParent
	def	viewGoToParent(sender)
		setValue_forKey(@source.classParent(@rightClickedLayer), 'value')
	end
	ib_action :viewGoToNextSibling
	def viewGoToNextSibling(sender)
		setValue_forKey(@source.classNextSibling(@rightClickedLayer), 'value')
	end
	ib_action :viewGoToPreviousSibling
	def viewGoToPreviousSibling(sender)
		setValue_forKey(@source.classPreviousSibling(@rightClickedLayer), 'value')
	end
	def viewGoToChild(sender)
		setValue_forKey(sender.title, 'value')
	end

	
	ib_action	:zoomActionFromSegmentedControl
	def	zoomActionFromSegmentedControl (sender)
		zoomOut		if sender.selectedSegment == 0
		zoomExtent	if sender.selectedSegment == 1
		zoomIn		if sender.selectedSegment == 2
	end
	
	def	zoomIn
		@frameworksLayerZ = [@frameworksLayerZ+@zoomAmount, @cameraMaximumZPosition].min

		CATransaction.begin
		CATransaction.setValue_forKey(0.8, KCATransactionAnimationDuration)
		updateFrameworksLayerTransform(@frameworksLayerX, @frameworksLayerY, @frameworksLayerZ)
		CATransaction.commit
	end

	def	zoomOut
		@frameworksLayerZ -= @zoomAmount

		CATransaction.begin
		CATransaction.setValue_forKey(0.8, KCATransactionAnimationDuration)
		updateFrameworksLayerTransform(@frameworksLayerX, @frameworksLayerY, @frameworksLayerZ)
		CATransaction.commit
	end

	def	zoomExtent
		CATransaction.begin
		CATransaction.setValue_forKey(0.8, KCATransactionAnimationDuration)
		layoutFrameworks
		zoomExtentLayers
		CATransaction.commit
	end

	# Called from toolbar's segmented control
	ib_action :toggleViewMode
	def toggleViewMode(sender)
		if sender.selectedSegment == 0
			switchToFrameworkView
		end
		if sender.selectedSegment == 1
			switchToClassHierarchyView
		end
	end
	

	def	switchToFrameworkView
		@viewMode = 'framework'
		
		deleteClassTree
		showFrameworkLayers(true)
	
		reflectSelectionInSelectionLayer
		zoomExtent
		
		updateColors if @colorsChangedSinceLastViewSwitch
		@colorsChangedSinceLastViewSwitch = false
	end
	
	def	switchToClassHierarchyView
		@viewMode = 'tree'

		deleteClassTree
		showFrameworkLayers(false)

		return	if @source == nil
		buildClassTree

		reflectSelectionInSelectionLayer
		zoomExtent		
		
		@colorsChangedSinceLastViewSwitch = false
		updateColors
	end
	
	def showFrameworkLayers(b)
		b = b ? 1 : 0
		@frameworksLayer.sublayers.each { |frameworkLayer|
			frameworkLayer.opacity = b if !frameworkLayer.class.to_s.match(/Text/) && frameworkLayer != @selectionLayer
		}
	
	end
	
	#
	# Build class tree
	#
	def buildClassTree

		# Recursive creation of layers
		def goDownStartingFrom(className, depth, y, yIncrement, parentLayer)
			layer = CATextLayer.layer
			layer.fontSize = @classLayerFontSize
			
			layer.position = CGPointMake(200*depth, y)
			layer.string = className
			layer.style = NSDictionary.dictionaryWithObjectsAndKeys('Futura-MediumItalic', 'font', KCAAlignmentCenter, 'alignmentMode', nil)
			s = layer.preferredFrameSize
			layer.bounds = CGRectMake(0, 0, s.width+layer.fontSize, s.height)
			layer.anchorPoint = CGPointMake(0, 0.5)
			layer.cornerRadius = 6

			@frameworksLayer.addSublayer layer
			
			# Draw a wire to parent layer
			if (parentLayer)
				wire = CALayer.layer
				wire.backgroundColor = @defaultFrameworkWireColor
				wire.anchorPoint = CGPointMake(1, 0.5)
				wire.position = CGPointMake(0, layer.bounds.size.height/2)
				wire.cornerRadius = 1
				layer.addSublayer wire
			
				vx = layer.position.x - (parentLayer.position.x + parentLayer.bounds.size.width)
				vy = layer.position.y - (parentLayer.position.y)
				len = Math.sqrt(vx*vx+vy*vy)
				vx /= len

				angle = len > 0 ? -Math.acos(vx) : 0

				wire.bounds = CGRectMake(0, 0, len, 2)
				wire.transform = CATransform3DMakeRotation(angle, 0, 0, 1)
			end
			
			y2 = y
			maxn = 0

			children = @source.hashChildren[className.to_s]
			children = [] if children == nil
			children.each { |child|
				n = goDownStartingFrom(child, depth+1, y2, yIncrement, layer)
				maxn += n
				y2 -= yIncrement*n
			}
			
			# Propagate maximum height reached by layouting children - used to offset parent's next siblings
			return [maxn, 1].max
		end
		
		roots = @source.rootObjects
		y = 0
		yIncrement = 17
		roots.each {	|root| 
						y -= goDownStartingFrom(root, 0, y*yIncrement, yIncrement, nil) }
	
	end
	
	def deleteClassTree
		layersToDelete = []
		@frameworksLayer.sublayers.each { |frameworkLayer|
			layersToDelete << frameworkLayer if frameworkLayer.class.to_s.match(/Text/)
		
		}
		layersToDelete.each { |frameworkLayer|
			frameworkLayer.removeFromSuperlayer
		}
	end


	
	#
	# Selected class
	#
	kvc_accessor :value
	def setValue_forKey(value, key)
		if key == 'value'
			return if value == @value
			self.performSelector_withObject_afterDelay(:updateWithNewValue, nil, 0)
		end
		
		super_setValue_forKey(value, key)
	end
	
	def updateWithNewValue
		reflectSelectionInSelectionLayer
	end
	
	
	#
	# The selection layer is overlaid on top of an existing class layer
	#
	def	createSelectionLayer
		
		layer = CALayer.layer
		layer.backgroundColor = CGColorCreateGenericRGB(0.9, 0.75, 0.2, 1)
#		layer.anchorPoint = CGPointMake(0, 0)
		layer.cornerRadius = @classLayerFontSize*0.7
		layer.contents = @image
		layer.masksToBounds = true
		layer.zPosition = 0.001
		layer.opacity = 0
		
		textLayer = CATextLayer.layer
		textLayer.fontSize = @classLayerFontSize
		textLayer.foregroundColor = CGColorCreateGenericRGB(0, 0, 0, 1)
		textLayer.position = CGPointMake(5, 0)
		textLayer.string = 'HELLO world'
		textLayer.style = NSDictionary.dictionaryWithObjectsAndKeys('Futura-MediumItalic', 'font', KCAAlignmentCenter, 'alignmentMode', nil)
		s = textLayer.preferredFrameSize
		textLayer.bounds = CGRectMake(0, 0, s.width+textLayer.fontSize, s.height)
		textLayer.anchorPoint = CGPointMake(0, 0)
		textLayer.cornerRadius = 5
		
		layer.bounds = CGRectMake(0, 0, s.width+textLayer.fontSize+10, s.height)

		layer.addSublayer textLayer
		@frameworksLayer.addSublayer layer
		
		@selectionLayer = layer
	end
	def	updateSelectionLayerWithSelectionText(text)
		textLayer = @selectionLayer.sublayers[0]
		textLayer.string = text

		s = textLayer.preferredFrameSize
		textLayer.bounds = CGRectMake(0, 0, s.width+textLayer.fontSize, s.height)
		@selectionLayer.bounds = CGRectMake(0, 0, s.width+textLayer.fontSize+10, s.height+2)
	end
	
	def reflectSelectionInSelectionLayer
		return	if !@loadedFrameworks

		# Don't animate move to position
		CATransaction.begin
		CATransaction.setValue_forKey(0, KCATransactionAnimationDuration)
		
		@selectedLayer = nil
		if @viewMode == 'framework'
			@frameworksLayer.sublayers.each { |frameworkLayer|
				if frameworkLayer.sublayers
					frameworkLayer.sublayers.each { |layer|
						if layer.class.to_s.match(/Text/)
							if layer.string == @value
								updateSelectionLayerWithSelectionText(@value)
								x = frameworkLayer.position.x+layer.position.x-frameworkLayer.bounds.size.width/2+layer.bounds.size.width/2-1
								y = frameworkLayer.position.y+layer.position.y-0*frameworkLayer.bounds.size.height+layer.bounds.size.height/2+1-@frameworkLayerStartHeight
								@selectionLayer.position = CGPointMake(x, y)
								@selectedLayer = layer
							end
						end
					}
				end
			}
		end
		if @viewMode == 'tree'
			@frameworksLayer.sublayers.each { |frameworkLayer|
				if frameworkLayer.class.to_s.match(/Text/) && frameworkLayer.string == @value
					updateSelectionLayerWithSelectionText(@value)
					x = frameworkLayer.position.x+frameworkLayer.bounds.size.width/2
					y = frameworkLayer.position.y
					@selectionLayer.position = CGPointMake(x, y)
					@selectedLayer = frameworkLayer
				end
			}
		end
		CATransaction.commit

		
		if @value.to_s.length > 0 && @selectedLayer
			anim = CAKeyframeAnimation.animationWithKeyPath('transform.scale')
			anim.values = [1, 1.2, 1]
			anim.timingFunctions = [CAMediaTimingFunction.functionWithName(KCAMediaTimingFunctionEaseInEaseOut), CAMediaTimingFunction.functionWithName(KCAMediaTimingFunctionEaseInEaseOut)]
			anim.duration = 0.15
			anim.repeatCount = 1
			@selectionLayer.addAnimation_forKey(anim, 'selectionUpdated')
		end
		hasSelection = @value.to_s.length > 0
		@selectionLayer.opacity = hasSelection ? 1 : 0
		@selectionLayer.backgroundColor = @isFirstResponder ? CGColorCreateGenericRGB(1, 0.8, 0.2, 1) : CGColorCreateGenericRGB(0.9, 0.75, 0.2, 1)
		
		scrollSelectionLayerToVisible if hasSelection 
	end
	
	#
	# Scroll to make selection layer visible
	#
	def	scrollSelectionLayerToVisible
	
		# Desired altitude
		z = [@zPositionWhenSelecting, @frameworksLayerZ].max

		# Instantly go to our desired altitude - we do that to compute our desired new coordinates
		CATransaction.begin
		CATransaction.setValue_forKey(0.0, KCATransactionAnimationDuration)
		updateFrameworksLayerTransform(@frameworksLayerX, @frameworksLayerY, z)
		CATransaction.commit
	
		# Convert selection layer rect in screen rect coordinates
		rect = CGRectMake(0, 0, @selectionLayer.bounds.size.width, @selectionLayer.bounds.size.height)
		screenRect = @selectionLayer.convertRect_toLayer(rect, self.layer)

		# The two boundary points of selection layer in screen coordinates
		p1 = CGPointMake(screenRect.origin.x, screenRect.origin.y)
		p2 = CGPointMake(screenRect.origin.x+screenRect.size.width, screenRect.origin.y+screenRect.size.height)

		# Check if rect is visible
		visible = p1.x > 0 && p2.x < frame.width && p1.y > 0 && p2.y < frame.height

		# If not, center it
		scale = frameworksLayerCoordinatesFromViewCoordinates(CGPointMake(1, 0)).x - frameworksLayerCoordinatesFromViewCoordinates(CGPointMake(0, 0)).x
		x = @frameworksLayerX + (frame.width/2 - (p1.x+p2.x)/2)*scale
		y = @frameworksLayerY + (frame.height/2 - (p1.y+p2.y)/2)*scale

		# Now that we have our coordinates, instantly go back to our original altitude
		CATransaction.begin
		CATransaction.setValue_forKey(0.0, KCATransactionAnimationDuration)
		updateFrameworksLayerTransform(@frameworksLayerX, @frameworksLayerY, @frameworksLayerZ)
		CATransaction.commit

		# If rect is visible and we are at a correct altitude, don't do anything
		return	if visible && @frameworksLayerZ >= @zPositionWhenSelecting
		# Setup new coordinates
		@frameworksLayerX = x
		@frameworksLayerY = y
		@frameworksLayerZ = z

		
		# And animate to them
		CATransaction.begin
		CATransaction.setValue_forKey(0.8, KCATransactionAnimationDuration)
		updateFrameworksLayerTransform(@frameworksLayerX, @frameworksLayerY, @frameworksLayerZ)
		CATransaction.commit
	end
	
	
	def isLayerAClassLayer(layer)
		clickedTextLayer = layer.class.to_s.match(/Text/)
		selection = nil
		if clickedTextLayer
			# Clicked a layer in tree view
			if layer.superlayer == @frameworksLayer
				selection = layer.string
			# Clicked a layer in framework view
			elsif layer.fontSize == @classLayerFontSize
				selection = layer.string
			# Clicked a framework layer
			else
				
			end
		end
		selection
	end
	def	maybeSelectLayer(layer)
		return	if layer == @selectionLayer

		selection = isLayerAClassLayer(layer)
		
		@selectedLayer = selection ? layer : nil
		selection = '' if selection == nil
		
		setValue_forKey(selection, 'value')
	end
	
	

	#
	#
	#	Colors
	#
	#
	def colorForFramework(frameworkPath)
		color = @source.colorForFramework(frameworkPath)
		return	@defaultFrameworkColor if color == nil
		color
	end
	def textColorForFramework(frameworkPath)
		color = @source.textColorForFramework(frameworkPath)
		return	@defaultFrameworkTextColor if color == nil
		color
	end
	
	

	def classLayers
		layers = []
		if @viewMode == 'framework'
			@frameworksLayer.sublayers.each { |frameworkLayer|
				next if frameworkLayer.class.to_s.match(/Text/)
				next if frameworkLayer == @selectionLayer || !frameworkLayer.sublayers
				frameworkLayer.sublayers.each { |layer|
					next if layer.fontSize != @classLayerFontSize
					layers << layer
				}
			}
		else
			@frameworksLayer.sublayers.each { |layer|
				next if !layer.class.to_s.match(/Text/)
				next if layer == @selectionLayer
				layers << layer
			}
		end
		layers
	end
	
	#
	# Given an input layer and a vector, find closest layer in vector direction
	#	Used to move with arrow keys in layers
	#
	def closestLayerFromLayerWithVector(inputLayer, x, y)
		return	if !inputLayer

		# Input layer position
		p0				= inputLayer.convertPoint_toLayer(CGPointMake(0, 0), self.layer)
		# 
		distance		= 1e99
		searchAngle		= Math.atan2(x, y)
		closestLayer	= nil

		# The layers can lie in that angle distance of the vector direction
		# From the input vector, we get an angle to match, 
		# maximumAngleDistance enlarges that angle to [-maximumAngleDistance/2, maximumAngleDistance/2]
		maximumAngleDistance = Math::PI/2
		
		# Check each layer
		classLayers.each { |layer|
			next if layer == inputLayer

			# Vector between inputLayer and current layer
			p = layer.convertPoint_toLayer(CGPointMake(0, 0), self.layer)
			vx = p.x-p0.x
			vy = p.y-p0.y

			# Normalize
			d = Math.sqrt(vx*vx + vy*vy)
			next if d < 1
			vx /= d
			vy /= d

			# Distance between the two angles
			angle = Math.atan2(vx, vy)
			angleDistance = (angle - searchAngle).abs % (2*Math::PI)
			angleDistance = 2*Math::PI - angleDistance if (angleDistance > Math::PI)
			
			# Record as closest layer
			if d < distance && angleDistance < maximumAngleDistance/2
				distance = d
				closestLayer = layer
			end
		}
		closestLayer
	end

	#
	# Given a vector, move to next layer in that direction
	# Keep selection if there's no next layer
	#
	def moveToNextLayerInDirection(x, y)
		closestLayer = closestLayerFromLayerWithVector(@selectedLayer, x, y)
		maybeSelectLayer(closestLayer) if closestLayer
	end	

	#
	# First responder methods
	#	We need them to handle arrow keys to move around the classes (moveDown, moveLeft, …)
	#
	def	acceptsFirstResponder
		true
	end
	
	def becomeFirstResponder
		@isFirstResponder = true
		return	super_becomeFirstResponder
	end
	
	def resignFirstResponder
		@isFirstResponder = false
		return	super_resignFirstResponder
	end
	

	#
	# Responder methods for arrow keys : select closest layer in that direction
	#
	def	moveDown(event)
		moveToNextLayerInDirection(0, -1)
	end
	def	moveLeft
		moveToNextLayerInDirection(-1, 0)
	end
	def	moveUp
		moveToNextLayerInDirection(0, 1)
	end
	def	moveRight
		moveToNextLayerInDirection(1, 0)
	end
	
	
	#
	# Color update
	#
	def updateColors
		if (@viewMode == 'framework')
			@frameworksLayer.sublayers.each { |fw|
				next if fw == @selectionLayer
				next if fw.class.to_s.match(/Text/)
				next if fw == @helperLayer
				frameworkPath = fw.sublayers[0].sublayers[0].string
				fw.backgroundColor = colorForFramework(frameworkPath)
				
				textColor = textColorForFramework(frameworkPath)
				fw.sublayers.each { |textLayer|
					textLayer.foregroundColor = textColor
				}
			}
		end
		if (@viewMode == 'tree')
			hashColors = {}
			hashTextColors = {}
			@frameworksLayer.sublayers.each { |fw|
				next if !fw.class.to_s.match(/Text/)
				frameworkPath = @source.frameworkOfClass(fw.string)
				hashColors[frameworkPath]		= colorForFramework(frameworkPath) if !hashColors[frameworkPath]
				hashTextColors[frameworkPath]	= textColorForFramework(frameworkPath) if !hashTextColors[frameworkPath]
				color = 
				
				fw.backgroundColor = hashColors[frameworkPath]
				fw.foregroundColor = hashTextColors[frameworkPath]
			}
		end
		
		@colorsChangedSinceLastViewSwitch = true
	end



	#
	# Full Screen
	#
	def toggleFullScreen(sender)
		if @isFullScreen
			exitFullScreen
			return
		end
		enterFullScreen
	end
	
	def	enterFullScreen
		r = enterFullScreenMode_withOptions(window.screen, nil)
		if r
			zoomExtent
			@isFullScreen = true
		end
	end
	
	def exitFullScreen
		r = exitFullScreenModeWithOptions(nil)
		@isFullScreen = false
		# Exiting fullscreen reinserts view as last child - reorder it
		# should also restore splitview's original frame sizes
		if superview.subviews[1].class.to_s.match(/Framework/)
			superview.addSubview superview.subviews[0]
		end

		self.performSelector_withObject_afterDelay(:zoomExtent, nil, 0)
		window.makeFirstResponder self
	end

	# When in fullscreen, handle esc — else dispatch to next responder
	def keyDown(event)
		if @isFullScreen && event.charactersIgnoringModifiers.hasPrefix("\033")
			exitFullScreen 
			return
		end
		nextResponder.keyDown(event)	if nextResponder
	end


end
