#
#  ItemTableView.rb
#  CocoaNav
#
#  Created by Patrick Geiller on 10/05/08.
#  Copyright (c) 2008 __MyCompanyName__. All rights reserved.
#

require 'osx/cocoa'

#
# ItemTableView uses two custom cells to draw data
#
class ItemTableView < NSTableView

	def awakeFromNib
	
		self.delegate = self
		self.setBackgroundColor(NSColor.whiteColor)
		
#		selectRowIndexes_byExtendingSelection(NSIndexSet.indexSet, false)
	end

	def tableView_shouldEditTableColumn_row(tableView, tableColumn, rowIndex)
		false
	end
	
	def	menuForEvent(event)
		row = rowAtPoint(convertPoint_fromView(event.locationInWindow, nil))
		selectRowIndexes_byExtendingSelection(NSIndexSet.indexSetWithIndex(row), false)
		return	self.menu
	end
	
	def selectRowIndexes_byExtendingSelection(indexes, extend)
		super_selectRowIndexes_byExtendingSelection(indexes, extend)
		
		return if @source == nil

		item = @source.itemList.arrangedObjects[selectedRow]
		className = ''
		className = item.name if item.oc_type == 'class'
		className = item.associatedClass if item.associatedClass != nil
		
		setValue_forKey(className, 'value')
	end
	
	def	rowHeight
		15
	end
	
	def setSource(source)
		@source = source
	end
	
	# Double click action
	def	doubleAction
		'openInFinder:'
	end


	#
	# Selected class
	#
	kvc_accessor :value
	def setValue_forKey(value, key)
#		puts 'from tableview : ' + value.to_s + '=' + key + '-'
		if key == 'value'
			return if value == @value
			self.performSelector_withObject_afterDelay(:updateWithNewValue, nil, 0)
		end
		super_setValue_forKey(value, key)
	end

	def updateWithNewValue
		return if selectedRow == -1
		items = @source.itemList.arrangedObjects
		item = items[selectedRow]

		# Don't do anything if an item associated to the class is selected
		return	if item.associatedClass == @value

		# Find class in controller objects
		items.each_with_index { |item, index|
			if item.oc_type == 'class' && item.name == @value
				selectRowIndexes_byExtendingSelection(NSIndexSet.indexSetWithIndex(index), false)
				scrollRowToVisible(index)
				return
			end
		}
	end


end


class ItemTableViewCell < NSTextFieldCell

	def awakeFromNib
		@shadow = NSShadow.alloc.init
		@shadow.shadowOffset = NSMakeSize(0, -1)
		@shadow.shadowBlurRadius = 4
		@shadow.shadowColor = NSColor.colorWithDeviceRed_green_blue_alpha(0, 0, 0, 0.75)
	end


	# Selected cell gets called twice, first time with object, second with a text cell — we need to skip that last one
	def setObjectValue(obj)
		begin
			obj.name
			@item = obj
		rescue
		end
	end

	def drawInteriorWithFrame_inView(cellFrame, controlView)
		return if @item == nil

		shadow = nil
		if isHighlighted
			brightness = 1
			shadow = @shadow
		else
			brightness = 0		
		end

		cellFrame.x += 2
		if @item.oc_type == 'framework'
			fontSize = 12
			attributes = NSDictionary.dictionaryWithObjectsAndKeys(NSColor.colorWithDeviceRed_green_blue_alpha(brightness, brightness, brightness, 1), NSForegroundColorAttributeName,
																NSFont.boldSystemFontOfSize(fontSize), NSFontAttributeName,
																shadow, NSShadowAttributeName,
													nil);

			str = NSString.stringWithString(@item.name)
			delta = (cellFrame.size.height - str.sizeWithAttributes(attributes).height)/2 + 0.5
			cellFrame.y += delta
			str.drawInRect_withAttributes(cellFrame, attributes)

		elsif @item.oc_type == 'class'
			fontSize = 10
			attributes = NSDictionary.dictionaryWithObjectsAndKeys(NSColor.colorWithDeviceRed_green_blue_alpha(brightness, brightness, brightness, 0.9), NSForegroundColorAttributeName,
																NSFont.boldSystemFontOfSize(fontSize), NSFontAttributeName,
																shadow, NSShadowAttributeName,
													nil);

			str = NSString.stringWithString(@item.name)
			delta = (cellFrame.size.height - str.sizeWithAttributes(attributes).height)/2 + 0.5
			cellFrame.y += delta
			str.drawInRect_withAttributes(cellFrame, attributes)

		elsif @item.oc_type == 'method'
			fontSize = 10
			attributes = NSDictionary.dictionaryWithObjectsAndKeys(NSColor.colorWithDeviceRed_green_blue_alpha(brightness, brightness, brightness, 0.5), NSForegroundColorAttributeName,
																NSFont.boldSystemFontOfSize(fontSize), NSFontAttributeName,
																shadow, NSShadowAttributeName,
													nil);

			# class name
			str = NSString.stringWithString(@item.associatedClass.to_s + ' ' + (@item.isInstanceMethod == true ? '':'+') + ' ')
			delta = (cellFrame.size.height - str.sizeWithAttributes(attributes).height)/2 + 0.5
			cellFrame.y += delta
			str.drawInRect_withAttributes(cellFrame, attributes)
			cellFrame.x += str.sizeWithAttributes(attributes).width

			attributes = NSDictionary.dictionaryWithObjectsAndKeys(NSColor.colorWithDeviceRed_green_blue_alpha(brightness, brightness, brightness, 0.8), NSForegroundColorAttributeName,
																NSFont.boldSystemFontOfSize(fontSize), NSFontAttributeName,
																shadow, NSShadowAttributeName,
													nil);
			# method name
			str = @item.name
			str.drawInRect_withAttributes(cellFrame, attributes)

		elsif @item.oc_type == 'category' || @item.oc_type == 'protocol'
			fontSize = 10
			# 'category ' or 'protocol '
			attributes = NSDictionary.dictionaryWithObjectsAndKeys(NSColor.colorWithDeviceRed_green_blue_alpha(brightness, brightness, brightness, 0.3), NSForegroundColorAttributeName,
																NSFont.boldSystemFontOfSize(fontSize), NSFontAttributeName,
																shadow, NSShadowAttributeName,
													nil);

			str = NSString.stringWithString(@item.oc_type.to_s + ' ')
			delta = (cellFrame.size.height - str.sizeWithAttributes(attributes).height)/2 + 0.5
			cellFrame.y += delta
			str.drawInRect_withAttributes(cellFrame, attributes)
			cellFrame.x += str.sizeWithAttributes(attributes).width

			# name
			attributes = NSDictionary.dictionaryWithObjectsAndKeys(NSColor.colorWithDeviceRed_green_blue_alpha(brightness, brightness, brightness, 0.8), NSForegroundColorAttributeName,
																NSFont.boldSystemFontOfSize(fontSize), NSFontAttributeName,
																shadow, NSShadowAttributeName,
													nil);
			str = @item.name
			str.drawInRect_withAttributes(cellFrame, attributes)
			cellFrame.x += str.sizeWithAttributes(attributes).width

			# for
			attributes = NSDictionary.dictionaryWithObjectsAndKeys(NSColor.colorWithDeviceRed_green_blue_alpha(brightness, brightness, brightness, 0.3), NSForegroundColorAttributeName,
																NSFont.boldSystemFontOfSize(fontSize), NSFontAttributeName,
																shadow, NSShadowAttributeName,
													nil);
			str = NSString.stringWithString(' for ')
			str.drawInRect_withAttributes(cellFrame, attributes)
			cellFrame.x += str.sizeWithAttributes(attributes).width

			# associated class
			attributes = NSDictionary.dictionaryWithObjectsAndKeys(NSColor.colorWithDeviceRed_green_blue_alpha(brightness, brightness, brightness, 0.66), NSForegroundColorAttributeName,
																NSFont.boldSystemFontOfSize(fontSize), NSFontAttributeName,
																shadow, NSShadowAttributeName,
													nil);
			str = @item.associatedClass
			str.drawInRect_withAttributes(cellFrame, attributes)
			cellFrame.x += str.sizeWithAttributes(attributes).width
		end
	end
end



class ItemTableViewSmallTextCell < NSTextFieldCell
	def drawInteriorWithFrame_inView(cellFrame, controlView)
		if isHighlighted
			color = NSColor.colorWithDeviceRed_green_blue_alpha(1, 1, 1, 0.7)
		else
			color = NSColor.colorWithDeviceRed_green_blue_alpha(0, 0, 0, 0.7)
		end

		attributes = NSDictionary.dictionaryWithObjectsAndKeys(color, NSForegroundColorAttributeName,
																NSFont.menuFontOfSize(10), NSFontAttributeName,
																nil);

		# Convert string to Ruby string
		str = stringValue.to_s
		# Keep only last path item
		str = str.split('/')[-1] if str.match(/\//)
		# Remove '.framework'
		str = str.split('.framework')[0] if str.match(/\.framework/)
		
		# Convert string back to ObjC
		str = NSString.stringWithString(str)
		# Center vertically
		size = str.sizeWithAttributes(attributes)
		delta = (cellFrame.size.height - size.height)/2 + 0.5
		cellFrame.origin.y += delta
		str.drawInRect_withAttributes(cellFrame, attributes)
	end

end
